#include <iostream>
#include <thread>
#include <vector>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <netinet/in.h>
#include <cstring>
#include <unistd.h>

#define PORT 8080
#define BACKLOG 10
#define BUFFER_SIZE 1024

void initOpenSSL() {
    SSL_load_error_strings();   
    OpenSSL_add_ssl_algorithms();
}

SSL_CTX* createContext() {
    SSL_CTX* ctx = SSL_CTX_new(TLS_server_method());
    if (!ctx) {
        ERR_print_errors_fp(stderr);
        exit(EXIT_FAILURE);
    }
    return ctx;
}

void configureContext(SSL_CTX* ctx) {
    SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv2);
    if (SSL_CTX_use_certificate_file(ctx, "cert.pem", SSL_FILETYPE_PEM) <= 0 ||
        SSL_CTX_use_PrivateKey_file(ctx, "key.pem", SSL_FILETYPE_PEM) <= 0) {
        ERR_print_errors_fp(stderr);
        exit(EXIT_FAILURE);
    }
}

std::string handleRequest(const std::string& request) {
    std::cout << "Received request:\n" << request << std::endl;
    return "HTTP/1.1 200 OK\n\
Content-Length: 13\n\
\n\
Hello, world!";
}

void handleClient(SSL* ssl) {
    char buffer[BUFFER_SIZE] = {0};
    
    // Accept SSL connection
    if (SSL_accept(ssl) <= 0) {
        ERR_print_errors_fp(stderr);
        return;
    }

    // Read from the SSL socket
    SSL_read(ssl, buffer, sizeof(buffer)); 
    std::string request(buffer);
    std::string response = handleRequest(request);

    // Write the response
    SSL_write(ssl, response.c_str(), response.size()); 

    SSL_shutdown(ssl); 
    SSL_free(ssl); 
}

void webServer() {
    initOpenSSL();
    SSL_CTX* ctx = createContext();
    configureContext(ctx);

    int server_fd;
    struct sockaddr_in address;
    int opt = 1;
    int addrlen = sizeof(address);

    if ((server_fd = socket(AF_INET, SOCK_STREAM, 0)) == 0) {
        perror("Socket failed");
        exit(EXIT_FAILURE);
    }

    if (setsockopt(server_fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt))) {
        perror("setsockopt");
        exit(EXIT_FAILURE);
    }

    address.sin_family = AF_INET;
    address.sin_addr.s_addr = INADDR_ANY;
    address.sin_port = htons(PORT);

    if (bind(server_fd, (struct sockaddr *)&address, sizeof(address)) < 0) {
        perror("Bind failed");
        exit(EXIT_FAILURE);
    }

    if (listen(server_fd, BACKLOG) < 0) {
        perror("Listen failed");
        exit(EXIT_FAILURE);
    }

    std::cout << "Server is listening on port " << PORT << std::endl;

    while (true) {
        int new_socket = accept(server_fd, (struct sockaddr *)&address, (socklen_t*)&addrlen);
        if (new_socket < 0) {
            perror("Accept failed");
            continue;
        }

        // Create a new SSL object for the connection
        SSL* ssl = SSL_new(ctx);
        SSL_set_fd(ssl, new_socket); // Bind the SSL object to the socket

        // Create a new thread to handle the client
        std::thread(handleClient, ssl).detach(); // Используем detach, чтобы поток продолжал работать независимо
    }

    SSL_CTX_free(ctx);
    EVP_cleanup();
}

int main() {
    webServer();
    return 0;
}
